/*  GnomeKiss - A KiSS viewer for the GNOME desktop
    Copyright (C) 2000  Nick Lamb <njl195@zepler.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include "kiss.h"

/* Concrete instantiation */
GSList *cells= NULL;

#define NIBBLE(x) { if (j & 1) x = (*(tmp++)) & 15; else x = (*tmp) >> 4; }

static int cell_load(KissCell *cell, guchar *palette, char *filename) {
  FILE *fp;
  guchar header[32], *buffer, *tmp;
  guchar k;
  int bpp, i, j;

  fp= open_any(filename, "r");
  if (fp == NULL) return 1;

  fread(header, 4, 1, fp);

  if (strncmp(header, "KiSS", 4)) {
    bpp= 4;
    cell->width= header[0] + (256 * header[1]);
    cell->height= header[2] + (256 * header[3]);
    cell->offx= 0;
    cell->offy= 0;
  } else { /* New-style image file, read full header */
    fread(header + 4, 28, 1, fp);
    bpp= header[5];
    cell->width= header[8] + (256 * header[9]);
    cell->height= header[10] + (256 * header[11]);
    cell->offx= header[12] + (256 * header[13]);
    cell->offy= header[14] + (256 * header[15]);
  }

  if (bpp < 32 && !palette) {
    g_message("No palette defined for \"%s\"", filename);
    return 1;
  }

  for (i= 0; i < SETS; ++i) {
    cell->x[i]= cell->offx;
    cell->y[i]= cell->offy;
  }

  cell->rgb= g_new(guchar, cell->width * cell->height * 3);
  cell->mask= g_new0(guchar, cell->width * cell->height);

  switch (bpp) {
  case 4:
    buffer= g_new(guchar, (cell->width + 1) / 2);
    for (i= 0; i < cell->height; ++i) {
      fread(buffer, (cell->width + 1) / 2, 1, fp);
      for (j= 0, tmp= buffer; j < cell->width; ++j) {
        NIBBLE(k);
        cell->mask[i * cell->width + j]= k ? 255 : 0;
        cell->rgb[3 * (i * cell->width + j)] = palette[k * 3];
        cell->rgb[3 * (i * cell->width + j) + 1] = palette[k * 3 + 1];
        cell->rgb[3 * (i * cell->width + j) + 2] = palette[k * 3 + 2];
      }
    }
    g_free(buffer);
    break;
  case 8:
    buffer= g_new(guchar, cell->width);
    for (i= 0; i < cell->height; ++i) {
      fread(buffer, cell->width, 1, fp);
      for (j= 0; j < cell->width; ++j) {
        cell->mask[i * cell->width + j]= (buffer[j]) ? 255 : 0;
        cell->rgb[3 * (i * cell->width + j)] = palette[buffer[j] * 3];
        cell->rgb[3 * (i * cell->width + j) + 1] = palette[buffer[j] * 3 + 1];
        cell->rgb[3 * (i * cell->width + j) + 2] = palette[buffer[j] * 3 + 2];
      }
    }
    g_free(buffer);
    break;
  case 32:
    buffer= g_new(guchar, cell->width * 4);
    for (i= 0; i < cell->height; ++i) {
      fread(buffer, cell->width, 4, fp);
      for (j= 0; j < cell->width; ++j) {
        cell->mask[i * cell->width + j]= buffer[j*4 + 3];
        cell->rgb[3 * (i * cell->width + j) ]= buffer[j*4 + 2];
        cell->rgb[3 * (i * cell->width + j) + 1]= buffer[j*4 + 1];
        cell->rgb[3 * (i * cell->width + j) + 2]= buffer[j*4 ];
      }
    }
    g_free(buffer);
    break;
  default:
    for (i= 0; i < cell->height; ++i) {
      for (j= 0; j < cell->width; ++j) {
        cell->mask[i * cell->width + j]= 255 ;
        cell->rgb[3 * (i * cell->width + j)] = j % 255;
        cell->rgb[3 * (i * cell->width + j) + 1] = i % 255;
        cell->rgb[3 * (i * cell->width + j) + 2] = j % 255;
      }
    }
    g_message("%s\nuses an unsupported bitdepth", filename);
    fprintf(stderr, "Can't do %d bpp\n", bpp);
    break;
  }

  fclose(fp);
  cell->mapped= 1; /* mapped */

  /* Some idiots say (640,480) and then load an 800x600 background CEL */
  if (rgb_buf_width < cell->width + cell->offx) {
    rgb_buf_width= cell->width + cell->offx;
    rgb_buf_stride= rgb_buf_width * 3;
  }
  if (rgb_buf_height < cell->height + cell->offy) {
    rgb_buf_height= cell->height + cell->offy;
  }

  cell->name= g_strdup(filename);

  return 0;
}

void object_move(KissObject *object, int x, int y) {
  GSList *list;
  KissCollision *collision;

  object_set_location(object, x, y);

  for (list= object->collisions; list != NULL; list= g_slist_next(list)) {
    collision= (KissCollision *) list->data;

    if (cell_collision(collision->cell1, collision->cell2)) {
      if (!collision->together) {
        collision->together= 1;
        events(collision->collide);
      }
    } else {
      if (collision->together) {
        collision->together= 0;
        events(collision->apart);
      }
    }
  }
}

void object_set_location(KissObject *object, int x, int y) {
  GSList *list;
  KissCell *cell;

  if (x + object->width > rgb_buf_width)
    x= rgb_buf_width - object->width;
  if (y + object->height > rgb_buf_height)
    y= rgb_buf_height - object->height;

  for (list= object->cells; list != NULL; list= g_slist_next(list)) {
    cell= (KissCell *) list->data;
    cell->x[view] = x + cell->offx;
    cell->y[view] = y + cell->offy;
  }
  object->x[view]= x; object->y[view]= y;

}

void cell_render(gpointer data, gpointer user_data) {
  int alpha;
  KissCell *cell;
  guchar *src, *dest;
  int row, col;
  
  cell= (KissCell *) data;
  if (!cell->mapped || !cell->alpha || !(cell->visible & (1 << view)))
    return;
  for (row= 0; row < cell->height; ++row) {
    src= cell->rgb + (row * cell->width * 3);
    dest= rgb_buf + ((row + cell->y[view]) * rgb_buf_stride
                     + (cell->x[view] * 3));
    if (dest < rgb_buf)
      continue; /* Don't write so early */
    for (col= 0; col < cell->width; ++col) {
      if (cell->mask[row * cell->width + col] == 255 && cell->alpha == 255) {
        *(dest++)= *(src++);
        *(dest++)= *(src++);
        *(dest++)= *(src++);
      } else if (cell->mask[row * cell->width + col] == 0) {
        dest += 3;
        src +=3;
      } else {
        alpha = (cell->mask[row * cell->width + col] * cell->alpha) / 255;
        *dest= ((*(src++) * alpha) + (*dest * (255 - alpha))) / 256; dest++;
        *dest= ((*(src++) * alpha) + (*dest * (255 - alpha))) / 256; dest++;
        *dest= ((*(src++) * alpha) + (*dest * (255 - alpha))) / 256; dest++;
      }
    }
  }
}

/* NB Returns a pointer to the CELL */
KissCell *cell_intersect(KissCell *cell, int x, int y) {

  x-= cell->x[view];
  y-= cell->y[view];

  if (!cell->mapped || !cell->alpha || !(cell->visible & (1 << view)))
    return NULL;
  
  if (x < 0 || x >= cell->width)
    return NULL;

  if (y < 0 || y >= cell->height)
    return NULL;

  if (cell->mask[x + y * cell->width] == 0)
    return NULL;

  return cell;
}

KissCell *cell_find(gchar *name) {
  GSList *list;
  KissCell *current;

  for (list= cells; list != NULL; list= g_slist_next(list)) {
    current= (KissCell *) list->data;
    if (!g_strcasecmp(current->name, name))
      return current;
  }

  fprintf(stderr, "No cell matching \"%s\"\n", name);
  return NULL;
}

KissCell *cell_new(KissObject *object, guchar *pal, gchar *filename) {
  KissCell *cell;

  cell= g_new0(KissCell, 1);
  cell->object= object;
  if (cell_load(cell, pal, filename)) {
    g_free(cell);
    return NULL;
  }

  /* Add to object and global display lists */
  cells = g_slist_prepend(cells, cell);
  object->cells = g_slist_prepend(object->cells, cell);
  if (cell->width + cell->offx > object->width)
    object->width = cell->width + cell->offx;
  if (cell->height + cell->offy > object->height)
    object->height = cell->height + cell->offy;

  return cell;
}

void cell_delete(KissCell *cell) {
  GSList *list;

  g_free(cell->rgb);
  g_free(cell->mask);
  g_free(cell->name);

  for (list= cell->press; list != NULL; list= g_slist_next(list)) {
    g_free((char *) list->data);
  }
  for (list= cell->release; list != NULL; list= g_slist_next(list)) {
    g_free((char *) list->data);
  }

  g_free(cell);
}

