/*  GnomeKiss - A KiSS viewer for the GNOME desktop
    Copyright (C) 2000  Nick Lamb <njl195@zepler.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <string.h>
#include <ctype.h>
#include "kiss.h"

KissObject *object= NULL;
int rgb_buf_width, rgb_buf_height, rgb_buf_stride;
guchar *rgb_buf= NULL;
int view;

static int obj, numpal, views;
guchar *palettes[PALETTES]; /* concrete */

static void parse_size(guchar *entry);
static void parse_move(guchar *entry);
static void parse_palette(guchar *entry);
static void parse_cell(guchar *entry);

void parse_file(const char *filename) {
  FILE *fp;
  gchar *dirname;
  guchar line[800];
  int k, script_flag= 0;

  clean_up(); /* clean up anything from previous configs */

  fp= open_any(filename, "r");

  if (fp == NULL) return; /* abort */

  dirname= g_dirname(filename);
  chdir(dirname);
  g_free(dirname);

  /* create an appropriately sized array of objects */
  object= g_new0(KissObject, OBJECTS);

  views= 0; obj= 0; numpal= 0; /* re-initialise walk-through counters */
  script= ERRORS; /* write to errors by default again */

  /* reset background color */
  // hasbgcolor= 0;

  rgb_buf_width= 640;
  rgb_buf_height= 480;  /* re-initialise RGB buffer size */
  rgb_buf_stride= 1920;

  view= -1; /* re-initialise group view thingy */
fprintf(stderr, "Parsing \"%s\" configuration...\n", filename);
  while (!feof(fp)) {
    if (fgets(line, 800, fp) == NULL) break;
    switch (*line) {
    case '%':
      parse_palette(line);
      break;
    case '(':
      parse_size(line);
      break;
    case ';':
      if (line[1] == '@' && script_flag)
        parse_action(line+2);
      else if (!strncmp(line, ";@EventHandler", 14))
        script_flag= 1;
      else
        ; /* This is a comment */
      break;
    case '#':
      parse_cell(line);
      break;
    case '$':
    case ' ':
      parse_move(line);
      break;
    default:
    }
  }

  fclose(fp);
fprintf(stderr, "...done\n");

  /* If there were no palettes, add a pseudo-palette */
  if (numpal == 0) {
    palettes[numpal++]= g_new(guchar, 256 * 3);
    palettes[0][0]= 255;
    palettes[0][1]= 255;
    palettes[0][2]= 255;
  }

  rgb_buf = g_new0(guchar, rgb_buf_stride * rgb_buf_height);
  gtk_window_set_default_size(GTK_WINDOW(app), rgb_buf_width, rgb_buf_height);
  gtk_widget_set_usize(area, rgb_buf_width, rgb_buf_height);

  /* Sort out set visibility */

  for (k= SETS - 1; k >= 0; --k) {
    if (views & (1 << k)) {
      gtk_widget_set_sensitive(items[k], 1);
      gtk_widget_set_sensitive(buttons[k], 1);
      view= k;
    } else {
      gtk_widget_set_sensitive(items[k], 0);
      gtk_widget_set_sensitive(buttons[k], 0);
    }
  }
  
  events(initialize);
  switch_view(view);
}

/* parse_move */
/* FIXME: This is an evil hack, and should eventually be replaced */

static void parse_move(guchar *entry) {
  guchar *token;
  int x, y;

  if (*entry == '$') {
    obj= 0;
    entry+= 2;
    view++;
  }

  token= entry;
  for (token= entry; *token; token++) {
    if (*token == '*') {
      obj++;
    } else if (isdigit(*token) && sscanf(token, "%d,%d", &x, &y) == 2) {
      object_set_location(&object[obj++], x, y);
      while (isdigit(*token)) ++token;
    }
  }
}

static void parse_size(guchar *entry) {
  int width= 640, height= 480;

  sscanf(entry, "(%u,%u)", &width, &height);

  rgb_buf_height= height;
  rgb_buf_width= width;
  rgb_buf_stride= width * 3;
}

static void parse_palette(guchar *entry) {
  char filename[32];
  guchar buffer[32], *palette;
  FILE *fp;
  int bits, cols, groups, k;

  sscanf(entry, "%% %20s ", filename);

  if ((fp= open_any(filename, "r")) == NULL) { /* Fake palette */
    palette= g_new(guchar, 256 * 3);
    for (k= 0; k < 256; ++k) {
      palette[k * 3]= k;
      palette[k * 3 + 1]= k;
      palette[k * 3 + 2]= k;
    }
    palettes[numpal++]= palette;
    return;
  }

  fread(buffer, 32, 1, fp);

  if (strncmp(buffer, "KiSS", 4)) {
     palette= g_new(guchar, 256 * 3);
     palettes[numpal++]= palette;
     for (k= 0; k < 16; ++k) {
       palette[k * 3] = (buffer[k * 2] & 0xf0);
       palette[k * 3 + 1] = (buffer[k * 2 + 1] << 4);
       palette[k * 3 + 2] = (buffer[k * 2] << 4);
     }
  } else { /* New-style palette file */
    bits = buffer[5];
    cols = buffer[8] + 256 * buffer[9];
    groups= buffer[10] + 256 * buffer[11];
    palette= g_new(guchar, 256 * 3);
    palettes[numpal++]= palette;
    if (bits == 24) {
      fread(palette, cols, 3, fp);
    } else if (bits == 12) {
      fread(buffer, 32, 1, fp);
      for (k= 0; k < 16; ++k) {
        palette[k * 3] = (buffer[k * 2] & 0xf0);
        palette[k * 3 + 1] = (buffer[k * 2 + 1] << 4);
        palette[k * 3 + 2] = (buffer[k * 2] << 4);
      }
    }
  }

  fclose(fp);
}

static void parse_cell(guchar *entry) {
  int args, group= 0, stick= 0, vis= 0;
  unsigned int trans= 0;
  guchar *pal;
  char pair[10], filename[20], rest[64], *next;
  KissCell *cell;

  pal = palettes[0];
  args = sscanf(entry, "#%10[.0-9] %19s %63[*:\t 0-9] ;%%t%u",
                       pair, filename, rest, &trans);
  if (args < 2) return;

  if (args >= 3) {
    for (next= rest; (*next != '\0'); ++next) {
      if (*next == '*') {
        pal = palettes[atoi(++next)];
        while (isdigit(*next)) ++next; /* skip remaining digits */
      } else if (isdigit(*next)) {
        vis |= 1 << (atoi(next));
      }
    }
  }

  sscanf(pair, "%d.%d", &group, &stick);

  cell= cell_new(&object[group], pal, filename);

  if (cell == NULL) return; /* abort */

  cell->visible= vis ? vis : ((1 << SETS) - 1);
  cell->object->sticky+= stick;
  cell->alpha = 255 - trans;
  views |= cell->visible;

  return;
}
