/*  GnomeKiss - A KiSS viewer for the GNOME desktop
    Copyright (C) 2000  Nick Lamb <njl195@zepler.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include "kiss.h"

KissActionList initialize, errors, kalarms[ALARMS], ksets[SETS];

static void object_unmap(KissObject *object);
static void object_map(KissObject *object);
static void object_altmap(KissObject *object);
static void object_transparent(KissObject *object, int amount);
static void object_move_relative(KissObject *object, int x, int y);
static void play_midi(gchar *song);

static int control_flag;

static void object_unmap(KissObject *object) {
  GSList *list;

  for (list= object->cells; list != NULL; list= g_slist_next(list)) {
    ((KissCell *) list->data)->mapped= 0;
  }
}

static void object_map(KissObject *object) {
  GSList *list;

  for (list= object->cells; list != NULL; list= g_slist_next(list)) {
    ((KissCell *) list->data)->mapped= 1;
  }
}

static void object_altmap(KissObject *object) {
  GSList *list;

  for (list= object->cells; list != NULL; list= g_slist_next(list)) {
    ((KissCell *) list->data)->mapped^= 1;
  }
}

static void object_transparent(KissObject *object, int amount) {
  GSList *list;
  KissCell *cell;

  for (list= object->cells; list != NULL; list= g_slist_next(list)) {
    cell= (KissCell *) list->data;
    cell->alpha= CLAMP(cell->alpha - amount, 0, 255);
  }
}

static void object_move_relative(KissObject *object, int x, int y) {
  object_move(object, object->x[view] + x, object->y[view] + y);
}

static void play_midi(gchar *song) {
  gchar temp[128];
/* FIXME this is much too dangerous to be allowed in a polished version */
  sprintf(temp, "playmidi -a %s&", song);
  system(temp);
}

void events(GSList *list) {

  control_flag= 1;
  while (list != NULL) {
    trigger(list->data);
    list= g_slist_next(list);
  }
  render_all();
}

void trigger(KissAction *action) {
  int k;

  if (action == NULL) {
    fprintf(stderr, "NULL trigger :(\n");
    return;
  }

  if (action->type < CONTROL && control_flag == 0)
    return; /* FKiSS 3 if { ... } else { ... } endif */

  switch (action->type) {
  case ACTION_ALTMAP_CELL:
    KissCell(action->args[0])->mapped^=1;
    break;
  case ACTION_ALTMAP_OBJECT:
    object_altmap(KissObject(action->args[0]));
    break;
  case ACTION_CHANGESET:
    switch_view(KissValue(action->args[0]));
    break;
  case ACTION_CHANGECOL:
    fprintf(stderr, "Action changecol(...) unimplemented\n");
    break;
  case ACTION_DEBUG:
    gnome_app_warning(GNOME_APP(app), KissString(action->args[0]));
    fprintf(stderr, "DEBUG: %s\n", KissString(action->args[0]));
    break;
  case ACTION_MAP_CELL:
    KissCell(action->args[0])->mapped= 1;
    break;
  case ACTION_MAP_OBJECT:
    object_map(KissObject(action->args[0]));
    break;
  case ACTION_MOVE_OBJECT:
    object_move_relative(KissObject(action->args[0]),
          KissValue(action->args[1]), KissValue(action->args[2]));
    break;
  case ACTION_NOP:
    break;
  case ACTION_QUIT:
    fprintf(stderr, "WARNING: tried to quit\n");
    break;
  case ACTION_RANDOMTIMER:
    if (KissValue(action->args[2]) > 0)
      k= KissValue(action->args[1]) + (rand() % KissValue(action->args[2]));
    else
      k= KissValue(action->args[1]);
    set_timer(kalarms[KissValue(action->args[0])], k);
    break;
  case ACTION_SHELL:
    fprintf(stderr, "WARNING: tried to execute \"%s\"\n",
                                   KissString(action->args[0]));
    break;
  case ACTION_SOUND:
    gnome_sound_play(KissString(action->args[0]));
    break;
  case ACTION_TIMER:
    set_timer(kalarms[KissValue(action->args[0])], KissValue(action->args[1]));
    break;
  case ACTION_TRANSPARENT_CELL:
    k = KissCell(action->args[0])->alpha - KissValue(action->args[1]);
    KissCell(action->args[0])->alpha= CLAMP(k, 0, 255);
    break;
  case ACTION_TRANSPARENT_OBJECT:
    object_transparent(KissObject(action->args[0]), KissValue(action->args[1]));
    break;
  case ACTION_UNMAP_CELL:
    KissCell(action->args[0])->mapped= 0;
    break;
  case ACTION_UNMAP_OBJECT:
    object_unmap(KissObject(action->args[0]));
    break;
  case ACTION_VIEWPORT:
    fprintf(stderr, "Action viewport(...) ignored\n");
    break;
  case ACTION_WINDOW_SIZE:
    fprintf(stderr, "Action windowsize(...) ignored\n");
    break;

/* FKiSS 2 onwards */

  case ACTION_MOVE_BYX:
    object_move(KissObject(action->args[0]),
                KissObject(action->args[1])->x[view] +
                    KissValue(action->args[2]),
                KissObject(action->args[0])->y[view]);
    break;
  case ACTION_MOVE_BYY:
    object_move(KissObject(action->args[0]),
                KissObject(action->args[0])->x[view],
                KissObject(action->args[1])->y[view] +
                    KissValue(action->args[2]));
    break;
  case ACTION_MOVE_TO:
    object_move(KissObject(action->args[0]),
                KissValue(action->args[1]), KissValue(action->args[2]));
    break;
  case ACTION_MUSIC:
    play_midi(KissString(action->args[0]));
    break;
  case ACTION_NOTIFY:
    gnome_app_message(GNOME_APP(app), KissString(action->args[0]));
    break;

/* FKiSS 2.1 onwards */

  case ACTION_IF_FIXED:
    if (KissObject(action->args[0])->sticky)
      set_timer(kalarms[KissValue(action->args[1])],
                KissValue(action->args[2]));
    break;
  case ACTION_IF_MAPPED:
    if (KissCell(action->args[0])->mapped)
      set_timer(kalarms[KissValue(action->args[1])],
                KissValue(action->args[2]));
    break;
  case ACTION_IF_MOVED:
    fprintf(stderr, "Action ifmoved(...) unimplemented\n");
    break;
  case ACTION_IF_NOT_FIXED:
    if (!KissObject(action->args[0])->sticky)
      set_timer(kalarms[KissValue(action->args[1])],
                KissValue(action->args[2]));
    break;
  case ACTION_IF_NOT_MAPPED:
    if (!KissCell(action->args[0])->mapped)
      set_timer(kalarms[KissValue(action->args[1])],
                KissValue(action->args[2]));
  case ACTION_IF_NOT_MOVED:
    fprintf(stderr, "Action ifnotmoved(...) unimplemented\n");
    break;
  case ACTION_MOVE_RANDX:
    if (KissValue(action->args[2]) > 0)
      k= KissValue(action->args[1]) + (rand() % KissValue(action->args[2]));
    else
      k= KissValue(action->args[1]);
    object_move_relative(KissObject(action->args[0]), k, 0);
    break;
  case ACTION_MOVE_RANDY:
    if (KissValue(action->args[2]) > 0)
      k= KissValue(action->args[1]) + (rand() % KissValue(action->args[2]));
    else
      k= KissValue(action->args[1]);
    object_move_relative(KissObject(action->args[0]), 0, k);
    break;
  case ACTION_MOVE_TORAND:
    object_move(KissObject(action->args[0]), rand() % rgb_buf_width,
                                             rand() % rgb_buf_height);
    break;
  case ACTION_SETFIX:
    k= KissValue(action->args[1]);
    KissObject(action->args[0])->sticky= k;
    if (k == 0)
      events(KissObject(action->args[0])->unfix);
    break;

/* FKiSS 3.0 onwards */

   case ACTION_VAR_LET:
     KissValue(action->args[0]) = KissValue(action->args[1]);
     break;
   case ACTION_VAR_ADD:
     KissValue(action->args[0]) = KissValue(action->args[1])
                                + KissValue(action->args[2]);
     break;
   case ACTION_VAR_SUB:
     KissValue(action->args[0]) = KissValue(action->args[1])
                                - KissValue(action->args[2]);
     break;

/* FKiSS 3.0 control structures :( */

   case CONTROL_EQUAL:
     if (KissValue(action->args[0]) == KissValue(action->args[1]))
       control_flag= 1;
     else
       control_flag= 0;
     break;
   case CONTROL_NOT_EQUAL:
     if (KissValue(action->args[0]) != KissValue(action->args[1]))
       control_flag= 1;
     else
       control_flag= 0;
     break;
   case CONTROL_GREATER_THAN:
     if (KissValue(action->args[0]) > KissValue(action->args[1]))
       control_flag= 1;
     else
       control_flag= 0;
     break;
   case CONTROL_LESS_THAN:
     if (KissValue(action->args[0]) < KissValue(action->args[1]))
       control_flag= 1;
     else
       control_flag= 0;
     break;
   case CONTROL_ELSE:
     control_flag = (control_flag) ? 0 : 1;
     break;
   case CONTROL_ENDIF:
     control_flag = 1;
     break;

  default:
    fprintf(stderr, "Invalid action %d executed.\n", action->type);
  }

}
