/*  GnomeKiss - A KiSS viewer for the GNOME desktop
    Copyright (C) 2000  Nick Lamb <njl195@zepler.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include "kiss.h"

GSList **script;

static KissActionList *find_event(gchar *event, gchar *params);
static KissAction *new_action(KissActionType type);
static void *int_helper(gchar *integer);

static KissActionList *find_event(gchar *event, gchar *params) {
  KissCell *cell, *cell2;
  gchar cellname[33], celltwo[33];
  guint objid;

  /* EventHandler is dealt with elsewhere */
 
  if (!strcmp("alarm", event)) {
    return &kalarms[atoi(params)];

  } else if (!strcmp("apart", event)) { /* FKiSS 2.1 */
    if (sscanf(params, "\"%[^\"]\" , \"%[^\"]\" ", cellname, celltwo)) {
      cell= cell_find(cellname);
      cell2= cell_find(celltwo);
      if (cell == NULL || cell2 == NULL)
        return ERRORS; /* FIXME */
      else
        return &(collision_find(cell, cell2)->apart);
    } else {
      return ERRORS; /* FIXME */
    }

  } else if (!strcmp("begin", event)) {
    return &initialize;

  } else if (!strcmp("catch", event)) {
    if (sscanf(params, "#%3u", &objid)) {
      return &object[objid].catch;
    } else {
      return ERRORS; /* FIXME */
    }

  } else if (!strcmp("col", event)) {
    return ERRORS; /* FIXME */

  } else if (!strcmp("collide", event)) { /* FKiSS 2.1 */
    if (sscanf(params, "\"%[^\"]\" , \"%[^\"]\" ", cellname, celltwo)) {
      cell= cell_find(cellname);
      cell2= cell_find(celltwo);
      if (cell == NULL || cell2 == NULL)
        return ERRORS; /* FIXME */
      else
        return &(collision_find(cell, cell2)->collide);
    } else {
      return ERRORS; /* FIXME */
    }

  } else if (!strcmp("drop", event)) {
    if (sscanf(params, "#%3u", &objid)) {
      return &object[objid].drop;
    } else {
      return ERRORS; /* FIXME */
    }

  } else if (!strcmp("end", event)) {
    return ERRORS; /* FIXME */

  } else if (!strcmp("fixcatch", event)) {
    if (sscanf(params, "#%3u", &objid)) {
      return &object[objid].fixcatch;
    } else {
      return ERRORS; /* FIXME */
    }

  } else if (!strcmp("fixdrop", event)) {
    if (sscanf(params, "#%3u", &objid)) {
      return &object[objid].fixdrop;
    } else {
      return ERRORS; /* FIXME */
    }

  } else if (!strcmp("in", event)) { /* FKiSS 2 */
    return ERRORS; /* FIXME */

  } else if (!strcmp("initialize", event)) {
    return &initialize;

  } else if (!strcmp("label", event)) { /* FKiSS 3 */
    return ERRORS; /* FIXME */

  } else if (!strcmp("never", event)) {
    return ERRORS; /* FIXME */

  } else if (!strcmp("out", event)) { /* FKiSS 2 */
    return ERRORS; /* FIXME */

  } else if (!strcmp("press", event)) {
    if (sscanf(params, "\"%[^\"]", cellname)) {
      if ((cell= cell_find(cellname)) != NULL)
        return &cell->press;
      else
        return ERRORS;
    } else if (sscanf(params, "#%3u", &objid)) {
      return &object[objid].press;
    }

  } else if (!strcmp("release", event)) {
    if (sscanf(params, "\"%[^\"]", cellname)) {
      if ((cell= cell_find(cellname)) != NULL)
        return &cell->release;
      else
        return ERRORS;
    } else if (sscanf(params, "#%3u", &objid)) {
      return &object[objid].release;
    }

  } else if (!strcmp("set", event)) {
    if (sscanf(params, "%2u", &objid))
      return &ksets[objid];

  } else if (!strcmp("stillin", event)) { /* FKiSS 2 */
    return ERRORS; /* FIXME */

  } else if (!strcmp("stillout", event)) { /* FKiSS 2 */
    return ERRORS; /* FIXME */

  } else if (!strcmp("unfix", event)) {
    if (sscanf(params, "\"%[^\"]", cellname)) {
      if ((cell= cell_find(cellname)) != NULL)
        return &(cell->object->unfix);
      else
        return ERRORS;
    } else if (sscanf(params, "#%3u", &objid)) {
      return &object[objid].unfix;
    }

  } else if (!strcmp("version", event)) { /* FKiSS 2 */
    if (atoi(params) <= 4) {
      return &initialize;       /* We claim FKiSS 3 compliance */
    } else {
      return ERRORS;
    }

  }
  return NULL;
}

static KissAction *new_action(KissActionType type) {
  KissAction *action;
  
  action= g_new0(KissAction, 1);
  action->type= type;
  return action;
}

static void *int_helper(gchar *integer) {
  int *result;

  result= var_find(integer);
  if (*result == 0) {
    *result= atoi(integer);
  }
  return (void *) result;
}

void check_and_add(KissAction *action, int required) {

  if (required > 0 && action->args[0] == NULL) {
    g_free(action);
  } else if (required > 1 && action->args[1] == NULL) {
    g_free(action);
  } else if (required > 2 && action->args[2] == NULL) {
    g_free(action);
  } else {
    *script= g_slist_append(*script, action);
  }
}

void parse_action(guchar *text) {
  gchar method[17], params[33], rest[128];
  gchar first[33], second[33], third[33];
  int one, two;
  KissAction *action;
  GSList **event;
  int count;

  count = sscanf(text, " %16[A-Za-z](%32[^)]) %128[^;]", method, params, rest);

  if (count == 1) {
    count = sscanf(text, " %16[A-Za-z]() %128[^;]", method, rest) + 1;
    params[0]= 0;
  }

  if (count < 2) {
    return;
  }

  if (!strcmp("add", method)) {
    if (sscanf(params, "%[A-Z0-9] , %[A-Z0-9-] , %[A-Z0-9-]",
                               first, second, third)) {
      action= new_action(ACTION_VAR_ADD);
      action->args[0]= int_helper(first);
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("altmap", method)) {
    if (sscanf(params, "\"%[^\"]\"", first)) {
      action= new_action(ACTION_ALTMAP_CELL);
      action->args[0]= cell_find(first);
      check_and_add(action, 1);
    } else if (sscanf(params, "#%3u", &one)) {
      action= new_action(ACTION_ALTMAP_OBJECT);
      action->args[0]= (void *) &object[one];
      check_and_add(action, 1);
    }

  } else if (!strcmp("changeset", method)) {
    if (sscanf(params, "%[A-Z0-9]", first)) {
      action= new_action(ACTION_CHANGESET);
      action->args[0]= int_helper(params);
      check_and_add(action, 1);
    }

  } else if (!strcmp("changecol", method)) {
    if (sscanf(params, "%[A-Z0-9]", first)) {
      action= new_action(ACTION_CHANGECOL);
      action->args[0]= int_helper(params);
      check_and_add(action, 1);
    }

  } else if (!strcmp("debug", method)) {
    if (sscanf(params, "\"%[^\"]", first)) {
      action= new_action(ACTION_DEBUG);
      action->args[0]= (void *) g_strdup(first);
      check_and_add(action, 1);
    }

  } else if (!strcmp("ifequal", method)) {
    if (sscanf(params, "%[A-Z0-9-] , %[A-Z0-9-] ", first, second)) {
      action= new_action(CONTROL_EQUAL);
      action->args[0]= int_helper(first);
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    }

  } else if (!strcmp("iffixed", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9] , %[A-Z0-9] ", &one, second, third)) {
      action= new_action(ACTION_IF_FIXED);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("ifgreaterthan", method)) {
    if (sscanf(params, "%[A-Z0-9-] , %[A-Z0-9-] ", first, second)) {
      action= new_action(CONTROL_GREATER_THAN);
      action->args[0]= int_helper(first);
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    }

  } else if (!strcmp("ifmapped", method)) {
    if (sscanf(params, "\"%[^\"]\" , %[A-Z0-9] , %[A-Z0-9]",
                                        first, second, third)) {
      action= new_action(ACTION_IF_MAPPED);
      action->args[0]= (void *) cell_find(first);
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("ifmoved", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9] , %[A-Z0-9] ", &one, second, third)) {
      action= new_action(ACTION_IF_MOVED);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("ifnotfixed", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9] , %[A-Z0-9] ", &one, second, third)) {
      action= new_action(ACTION_IF_NOT_FIXED);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("ifnotmapped", method)) {
    if (sscanf(params, "\"%[^\"]\" , %[A-Z0-9] , %[A-Z0-9]",
                                        first, second, third)) {
      action= new_action(ACTION_IF_NOT_MAPPED);
      action->args[0]= (void *) cell_find(first);
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("ifnotmoved", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9] , %[A-Z0-9]", &one, second, third)) {
      action= new_action(ACTION_IF_NOT_MOVED);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("let", method)) {
    if (sscanf(params, "%[A-Z0-9] , %[A-Z0-9-]", first, second)) {
      action= new_action(ACTION_VAR_LET);
      action->args[0]= int_helper(first);
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    }

  } else if (!strcmp("map", method)) {
    if (sscanf(params, "\"%[^\"]\"", first)) {
      action= new_action(ACTION_MAP_CELL);
      action->args[0]= (void *) cell_find(first);
      check_and_add(action, 1);
    } else if (sscanf(params, "#%3u", &one)) {
      action= new_action(ACTION_MAP_OBJECT);
      action->args[0]= (void *) &object[one];
      check_and_add(action, 1);
    }

  } else if (!strcmp("movebyx", method)) {
    if (sscanf(params, "#%3u , #%3u , %[A-Z0-9-]", &one, &two, third)) {
      action= new_action(ACTION_MOVE_BYX);
      action->args[0]= (void *) &object[one];
      action->args[1]= (void *) &object[two];
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("movebyy", method)) {
    if (sscanf(params, "#%3u , #%3u , %[A-Z0-9-]", &one, &two, third)) {
      action= new_action(ACTION_MOVE_BYY);
      action->args[0]= (void *) &object[one];
      action->args[1]= (void *) &object[two];
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("move", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9-] , %[A-Z0-9-]", &one, second, third)) {
      action= new_action(ACTION_MOVE_OBJECT);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("moverandx", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9-] , %[A-Z0-9]", &one, second, third)) {
      action= new_action(ACTION_MOVE_RANDX);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("moverandy", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9-] , %[A-Z0-9]", &one, second, third)) {
      action= new_action(ACTION_MOVE_RANDY);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("moveto", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9] , %[A-Z0-9]", &one, second, third)) {
      action= new_action(ACTION_MOVE_TO);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("movetorand", method)) {
    if (sscanf(params, "#%3u", &one)) {
      action= new_action(ACTION_MOVE_TORAND);
      action->args[0]= (void *) &object[one];
      check_and_add(action, 1);
    }

  } else if (!strcmp("music", method)) {
    if (sscanf(params, "\"%[^\"]", first)) {
      action= new_action(ACTION_MUSIC);
      if (case_fixup(first))
        action->args[0]= (void *) g_strdup(first);
      check_and_add(action, 1);
    }

  } else if (!strcmp("nop", method)) {
    action= new_action(ACTION_NOP);
    check_and_add(action, 0);

  } else if (!strcmp("notify", method)) {
    if (sscanf(params, "\"%[^\"]", first)) {
      action= new_action(ACTION_NOTIFY);
      action->args[0]= (void *) g_strdup(first);
      check_and_add(action, 1);
    }

  } else if (!strcmp("quit", method)) {
    action= new_action(ACTION_QUIT);
    check_and_add(action, 0);

  } else if (!strcmp("randomtimer", method)) {
    if (sscanf(params,"%[A-Z0-9] , %[A-Z0-9] , %[A-Z0-9]",
                                        first, second, third)) {
      action= new_action(ACTION_RANDOMTIMER);
      action->args[0]= int_helper(first);
      action->args[1]= int_helper(second);
      action->args[2]= int_helper(third);
      check_and_add(action, 3);
    }

  } else if (!strcmp("setfix", method)) {
    if (sscanf(params, "#%3u , %[A-Z0-9]", &one, second)) {
      action= new_action(ACTION_SETFIX);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    }

  } else if (!strcmp("shell", method)) {
    action= new_action(ACTION_SHELL);
    action->args[0]= (void *) g_strdup(params);
    check_and_add(action, 1);

  } else if (!strcmp("sound", method)) {
    if (sscanf(params, "\"%[^\"]", first)) {
      action= new_action(ACTION_SOUND);
      if (case_fixup(first))
        action->args[0]= (void *) g_strdup(first);
      check_and_add(action, 1);
    }

  } else if (!strcmp("timer", method)) {
    if (sscanf(params,"%[A-Z0-9] , %[A-Z0-9]", first, second)) {
      action= new_action(ACTION_TIMER);
      action->args[0]= int_helper(first);
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    }

  } else if (!strcmp("transparent", method)) {
    if (sscanf(params, "\"%[^\"]\" , %[A-Z0-9-]", first, second)) {
      action= new_action(ACTION_TRANSPARENT_CELL);
      action->args[0]= (void *) cell_find(first);
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    } else if (sscanf(params, "#%3u , %[A-Z0-9-]", &one, second)) {
      action= new_action(ACTION_TRANSPARENT_OBJECT);
      action->args[0]= (void *) &object[one];
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    }

  } else if (!strcmp("unmap", method)) {
    if (sscanf(params, "\"%[^\"]\"", first)) {
      action= new_action(ACTION_UNMAP_CELL);
      action->args[0]= (void *) cell_find(first);
      check_and_add(action, 1);
    } else if (sscanf(params, "#%3u", &one)) {
      action= new_action(ACTION_UNMAP_OBJECT);
      action->args[0]= (void *) &object[one];
      check_and_add(action, 1);
    }

  } else if (!strcmp("viewport", method)) {
    if (sscanf(params, "%[A-Z0-9-] , %[A-Z0-9-]", first, second)) {
      action= new_action(ACTION_VIEWPORT);
      action->args[0]= int_helper(first);
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    }

  } else if (!strcmp("windowsize", method)) {
    if (sscanf(params, "%[A-Z0-9-] , %[A-Z0-9-]", first, second)) {
      action= new_action(ACTION_WINDOW_SIZE);
      action->args[0]= int_helper(first);
      action->args[1]= int_helper(second);
      check_and_add(action, 2);
    }

  } else {
    event= find_event(method, params);
    if (event == ERRORS) {
      fprintf(stderr, "ERROR: %s(%s)\n", method, params);
      script= ERRORS;
    } else if (event == NULL) {
      fprintf(stderr, "UNKNOWN: %s(%s)\n", method, params);
    } else {
      script= event;
    }
  }

  if (count == 3) {
    parse_action(rest);
  }
}
